import * as S from "@effect/schema/Schema";
import * as O from "effect/Option";
import * as Types from "effect/Types";

// --- METHOD ------------------------------------------------------------------

export type Method =
	| "GET"
	| "POST"
	| "PUT"
	| "PATCH"
	| "DELETE"
	;

// --- ROUTE INPUT -------------------------------------------------------------

export type RouteInputLiteral<K extends string> = K;

export type RouteInputSchema<K extends string, T> = readonly [name: K, schema: S.Schema<T, string>];

export type RouteInputToken =
	| RouteInputLiteral<string>
	| RouteInputSchema<string, any>
	;

// --- ROUTE -------------------------------------------------------------------

export enum RouteTokenType {
	Literal = "literal",
	Schema = "schema",
}

export interface RouteTokenLiteral<K extends string> {
	readonly type: RouteTokenType.Literal,
	readonly literal: K,
}

export interface RouteTokenSchema<K extends string, T> {
	readonly type: RouteTokenType.Schema;
	readonly name: K;
	readonly schema: S.Schema<T, string>;
}

export type RouteToken =
	| RouteTokenLiteral<string>
	| RouteTokenSchema<string, any>
	;

export function routeTokenLiteral<const K extends string>(literal: K): RouteTokenLiteral<K> {
	return Object.freeze<RouteTokenLiteral<K>>({
		type: RouteTokenType.Literal,
		literal,
	});
}

export function routeTokenSchema<const K extends string, const T>(name: K, schema: S.Schema<T, string>): RouteTokenSchema<K, T> {
	return Object.freeze<RouteTokenSchema<K, T>>({
		type: RouteTokenType.Schema,
		name,
		schema,
	});
}

// --- ROUTE FROM INPUT --------------------------------------------------------

export type RouteFromInput<I extends readonly RouteInputToken[]> = {
	readonly [K in keyof I]:
		I[K] extends RouteInputLiteral<infer _K> ? RouteTokenLiteral<_K>
		: I[K] extends RouteInputSchema<infer _K, infer _T> ? RouteTokenSchema<_K, _T>
		: never;
}

export type ParamFromRouteToken<T extends RouteToken> =
	T extends RouteTokenSchema<infer K, infer V> ? { readonly [_ in K]: V } : never;

export type ParamsFromRoute<I extends readonly RouteToken[]> =
	Types.Simplify<Types.UnionToIntersection<ParamFromRouteToken<I[number]>>>;

export function makeRoute<const RouteInput extends readonly RouteInputToken[]>(...tokens: RouteInput): RouteFromInput<RouteInput> {
	const names = new Set<string>();
	const ret = new Array<RouteToken>(tokens.length);

	for (let i = 0; i < tokens.length; ++i) {
		const token = tokens[i];
		if (typeof token === "string") {
			ret[i] = routeTokenLiteral(token);
		} else {
			const [name, schema] = token;
			if (names.has(name)) {
				throw new Error(`Duplicate URL parameter name in route: ${name}`);
			}
			ret[i] = routeTokenSchema(name, schema);
		}
	}

	return Object.freeze(ret) as RouteFromInput<RouteInput>;
}

// --- API ---------------------------------------------------------------------

export interface ApiBase {
	readonly requestBody: O.Option<S.Schema<any, any>>;
	readonly query: O.Option<S.Schema<any, { readonly [_: string]: string }>>;
	readonly responseBody: O.Option<S.Schema<any, any>>;

	readonly method: Method;
	readonly route: readonly RouteToken[];
}

export type RequestType<T extends ApiBase> = S.Schema.Type<O.Option.Value<T["requestBody"]>>;
export type RequestEncoded<T extends ApiBase> = S.Schema.Encoded<O.Option.Value<T["requestBody"]>>;
export type QueryType<T extends ApiBase> = S.Schema.Type<O.Option.Value<T["query"]>>;
export type QueryEncoded<T extends ApiBase> = S.Schema.Encoded<O.Option.Value<T["query"]>>;
export type ResponseType<T extends ApiBase> = S.Schema.Type<O.Option.Value<T["responseBody"]>>;
export type ResponseEncoded<T extends ApiBase> = S.Schema.Encoded<O.Option.Value<T["responseBody"]>>;
export type Params<T extends ApiBase> = ParamsFromRoute<T["route"]>;

// --- MAKE API ----------------------------------------------------------------

export interface MakeApiProps {
	readonly requestBody?: S.Schema<any, any>;
	readonly query?: S.Schema<any, any>;
	readonly responseBody?: S.Schema<any, any>;

	readonly method: Method;
	readonly route: readonly RouteInputToken[];
}

export interface MakeApi<P extends MakeApiProps> {
	readonly requestBody: P extends { readonly requestBody: S.Schema<infer A, infer I> } ? O.Some<S.Schema<A, I>> : O.None<never>;
	readonly query: P extends { readonly query: S.Schema<infer A, infer I> } ? O.Some<S.Schema<A, I>> : O.None<never>;
	readonly responseBody: P extends { readonly responseBody: S.Schema<infer A, infer I> } ? O.Some<S.Schema<A, I>> : O.None<never>;

	readonly method: P["method"];
	readonly route: RouteFromInput<P["route"]>;
}

export function make<const P extends MakeApiProps>(props: P): MakeApi<P> {
	return Object.freeze<any>({
		requestBody: O.fromNullable(props.requestBody),
		query: O.fromNullable(props.query),
		responseBody: O.fromNullable(props.responseBody),

		method: props.method,
		route: makeRoute(...props.route),
	} satisfies ApiBase) as MakeApi<P>;
}
