/// <reference lib="dom" />
import * as Effect from "effect/Effect";
import { pipe } from "effect/Function";
import * as Option from "effect/Option";
import * as Predicate from "effect/Predicate";
import * as Stream from "effect/Stream";

export const FetchExceptionTypeId: unique symbol = Symbol.for("effect-api/Fetch/errors/FetchException");
export const RequestArrayBufferExceptionTypeId: unique symbol = Symbol.for("effect-api/Fetch/errors/RequestArrayBufferException");

export type FetchExceptionTypeId = typeof FetchExceptionTypeId;
export type RequestArrayBufferExceptionTypeId = typeof RequestArrayBufferExceptionTypeId;

export interface FetchException extends Error {
	readonly _tag: "FetchException"
	readonly [FetchExceptionTypeId]: FetchExceptionTypeId;
	readonly error: TypeError;
}

export interface RequestArrayBufferException extends Error {
	readonly _tag: "RequestArrayBufferException"
	readonly [RequestArrayBufferExceptionTypeId]: RequestArrayBufferExceptionTypeId;
	readonly error: RangeError;
}

const FetchExceptionProto = Object.create(Error.prototype, {
	_tag: { value: "FetchException" },
	[FetchExceptionTypeId]: { value: FetchExceptionTypeId },
});

const RequestArrayBufferExceptionProto = Object.create(Error.prototype, {
	_tag: { value: "RequestArrayBufferException" },
	[RequestArrayBufferExceptionTypeId]: { value: RequestArrayBufferExceptionTypeId },
});

export const FetchException: (error: TypeError) => FetchException = (error) => Object.create(FetchExceptionProto, {
	error: { value: error, enumerable: true },
});

export const RequestArrayBufferException: (error: RangeError) => RequestArrayBufferException = (error) => Object.create(RequestArrayBufferExceptionProto, {
	error: { value: error, enumerable: true },
});

export const isFetchException: (u: unknown) => u is FetchException = (u): u is FetchException => Predicate.hasProperty(u, FetchExceptionTypeId);
export const isRequestArrayBufferException: (u: unknown) => u is RequestArrayBufferException = (u): u is RequestArrayBufferException => Predicate.hasProperty(u, RequestArrayBufferExceptionTypeId);

const _wrapResponseHandler: ProxyHandler<Response> = Object.freeze<ProxyHandler<Response>>({
	get: (target, p, receiver) => {
		switch (p) {
			case "body":
				return <E>(onError: (error: unknown) => E): Option.Option<Stream.Stream<never, E, Uint8Array>> => {
					return pipe(
						Reflect.get(target, p, receiver),
						Option.fromNullable,
						Option.map(Stream.fromReadableStream())
					);
				};
		}
	}
});

const _wrapResponse: (response: Response) => _Response = (response) => {
};

const _fetch: (...args: Parameters<typeof fetch>) => Effect.Effect<never, FetchException, _Response> = (input, init) => {
	return pipe(
		Effect.tryPromise({
			try: (signal) => fetch(input, { signal, ...init }),
			catch: (error) => {
				if (error instanceof TypeError) {
					return FetchException(error);
				}
				throw error;
			},
		}),
		Effect.map(_wrapResponse),
	);
};

interface _Response {
	body<E>(onError: (error: unknown) => E): Option.Option<Stream.Stream<never, E, Uint8Array>>;
	readonly bodyUsed: boolean;
	arrayBuffer(): Effect.Effect<never, RequestArrayBufferException, ArrayBuffer>;
	blob(): Effect.Effect<never, ___, Blob>;
	formData(): Effect.Effect<never, ___, FormData>;
	json(): Effect.Effect<never, ___, unknown>;
	text(): Effect.Effect<never, ___, string>;
	readonly headers: Headers;
	readonly ok: boolean;
	readonly redirected: boolean;
	readonly status: number;
	readonly statusText: string;
	readonly type: ResponseType;
	readonly url: string;
	clone(): _Response;
}

export {
	_fetch as fetch,
	type _Request as Request
};

